=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2006-03-07 12:17:32 +0000
+++ bzrlib/knit.py	2006-03-08 02:12:22 +0000
@@ -791,25 +791,14 @@
                 for rec_id, pos, size in self.versions:
                     yield rec_id, fp.read(size)
 
-        fp = self._open_file()
-
-        # Loop through all records and try to collect as large
-        # continuous region as possible to read.
-        while records:
-            record_id, pos, size = records.pop(0)
-            continuous_range = ContinuousRange(record_id, pos, size)
-            while records:
-                record_id, pos, size = records[0]
-                if continuous_range.add(record_id, pos, size):
-                    del records[0]
-                else:
-                    break
-            fp.seek(continuous_range.start_pos, 0)
-            for record_id, data in continuous_range.split(fp):
-                content, digest = self._parse_record(record_id, data)
-                yield record_id, content, digest
-
-        self._file = None
+        # We take it that the transport optimizes the fetching as good
+        # as possible (ie, reads continous ranges.)
+        response = self._transport.readv(self._filename,
+            [(pos, size) for version_id, pos, size in records])
+
+        for (record_id, pos, size), (pos, data) in zip(records, response):
+            content, digest = self._parse_record(record_id, data)
+            yield record_id, content, digest
 
     def read_records(self, records):
         """Read records into a dictionary."""

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2006-02-28 07:42:57 +0000
+++ bzrlib/tests/test_transport_implementations.py	2006-03-08 02:12:22 +0000
@@ -892,3 +892,15 @@
         # TODO make this consistent on all platforms:
         # self.assertRaises(LockError, transport.lock_read, 'lock')
         lock.unlock()
+
+    def test_readv(self):
+        transport = self.get_transport()
+        if transport.is_readonly():
+            file('a', 'w').write('0123456789')
+        else:
+            transport.put('a', StringIO('01234567890'))
+
+        d = list(transport.readv('a', ((0, 1), (3, 1), (9, 1))))
+        self.assertEqual(d[0], (0, '0'))
+        self.assertEqual(d[1], (3, '3'))
+        self.assertEqual(d[2], (9, '9'))

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2006-02-28 07:42:57 +0000
+++ bzrlib/transport/__init__.py	2006-03-08 02:12:22 +0000
@@ -244,6 +244,17 @@
         """
         raise NotImplementedError
 
+    def readv(self, relpath, offsets):
+        """Get parts of the file at the given relative path.
+
+        :offsets: A list of (offset, size) tuples.
+        :return: A list or generator of (offset, data) tuples
+        """
+        fp = self.get(relpath)
+        for offset, size in offsets:
+            fp.seek(offset)
+            yield offset, fp.read(size)
+
     def get_multi(self, relpaths, pb=None):
         """Get a list of file-like objects, one for each entry in relpaths.
 

=== modified file 'bzrlib/transport/http.py'
--- bzrlib/transport/http.py	2006-02-23 01:46:22 +0000
+++ bzrlib/transport/http.py	2006-03-08 02:12:22 +0000
@@ -74,7 +74,7 @@
             return urllib2.Request.get_method(self)
 
 
-def get_url(url, method=None):
+def get_url(url, method=None, ranges=None):
     import urllib2
     mutter("get_url %s", url)
     manager = urllib2.HTTPPasswordMgrWithDefaultRealm()
@@ -85,6 +85,8 @@
     request = Request(url)
     request.method = method
     request.add_header('User-Agent', 'bzr/%s' % bzrlib.__version__)
+    if ranges:
+        request.add_header('Range', ranges)
     response = opener.open(request)
     return response
 
@@ -190,15 +192,11 @@
                 return False
             raise TransportError(orig_error=e)
 
-    def get(self, relpath, decode=False):
-        """Get the file at the given relative path.
-
-        :param relpath: The relative path to the file
-        """
+    def _get(self, relpath, decode=False, ranges=None):
         path = relpath
         try:
             path = self.abspath(relpath)
-            return get_url(path)
+            return get_url(path, ranges=ranges)
         except urllib2.HTTPError, e:
             mutter('url error code: %s for has url: %r', e.code, path)
             if e.code == 404:
@@ -214,6 +212,31 @@
                              % (self.abspath(relpath), str(e)),
                              orig_error=e)
 
+    def get(self, relpath, decode=False):
+        """Get the file at the given relative path.
+
+        :param relpath: The relative path to the file
+        """
+        return self._get(relpath, decode=decode)
+
+    def readv(self, relpath, offsets):
+        """Get parts of the file at the given relative path.
+
+        :offsets: A list of (offset, size) tuples.
+        :return: A list or generator of (offset, data) tuples
+        """
+        response = self._get(relpath,
+            ranges=','.join(['%d-%d' % (off, off + size - 1)
+                             for off, size in offsets]))
+        if response.code == 206:
+            for off, size in offsets:
+                yield off, response.read(size)
+        elif response.code == 200:
+            fp = StringIO(response.read())
+            for off, size in offsets:
+                fp.seek(off)
+                yield off, fp.read(size)
+
     def put(self, relpath, f, mode=None):
         """Copy the file-like or string object into the location.
 

